@@ -5,7 +5,9 @@ import (
 	"encoding/json"
 	"flag"
 	"fmt"
+	"os"
 	"reflect"
+	"strconv"
 	"strings"
 	"time"
 
@@ -27,7 +29,6 @@ import (
 	k8serrors "k8s.io/apimachinery/pkg/api/errors"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/runtime"
-	"k8s.io/apimachinery/pkg/types"
 
 	"sigs.k8s.io/controller-runtime/pkg/client"
 	"sigs.k8s.io/controller-runtime/pkg/controller"
@@ -40,16 +41,35 @@ import (
 )
 
 const (
-	hostErrorRetryDelay = time.Second * 10
+	hostErrorRetryDelay    = time.Second * 10
+	pauseRetryDelay        = time.Second * 30
+	rebootAnnotationPrefix = "reboot.metal3.io"
 )
 
 var runInTestMode bool
 var runInDemoMode bool
+var maxConcurrentReconciles int = 3
 
 func init() {
 	flag.BoolVar(&runInTestMode, "test-mode", false, "disable ironic communication")
 	flag.BoolVar(&runInDemoMode, "demo-mode", false,
 		"use the demo provisioner to set host states")
+
+	if mcrEnv, ok := os.LookupEnv("BMO_CONCURRENCY"); ok {
+		mcr, err := strconv.Atoi(mcrEnv)
+		if err != nil {
+			log.Error(err, fmt.Sprintf("BMO_CONCURRENCY value: %s is invalid", mcrEnv))
+			os.Exit(1)
+		}
+		if mcr > 0 {
+			log.Info(fmt.Sprintf("BMO_CONCURRENCY of %d is set via an environment variable", mcr))
+			maxConcurrentReconciles = mcr
+		} else {
+			log.Info(fmt.Sprintf("Invalid BMO_CONCURRENCY value. Operator Concurrency will be set to a default value of %d", maxConcurrentReconciles))
+		}
+	} else {
+		log.Info(fmt.Sprintf("Operator Concurrency will be set to a default value of %d", maxConcurrentReconciles))
+	}
 }
 
 var log = logf.Log.WithName("baremetalhost")
@@ -86,7 +106,9 @@ func newReconciler(mgr manager.Manager) reconcile.Reconciler {
 func add(mgr manager.Manager, r reconcile.Reconciler) error {
 	// Create a new controller
 	c, err := controller.New("metal3-baremetalhost-controller", mgr,
-		controller.Options{Reconciler: r})
+		controller.Options{MaxConcurrentReconciles: maxConcurrentReconciles,
+			Reconciler: r,
+		})
 	if err != nil {
 		return err
 	}
@@ -170,6 +192,14 @@ func (r *ReconcileBareMetalHost) Reconcile(request reconcile.Request) (result re
 		return reconcile.Result{}, errors.Wrap(err, "could not load host data")
 	}
 
+	// If the reconciliation is paused, requeue
+	annotations := host.GetAnnotations()
+	if annotations != nil {
+		if _, ok := annotations[metal3v1alpha1.PausedAnnotation]; ok {
+			return reconcile.Result{Requeue: true, RequeueAfter: pauseRetryDelay}, nil
+		}
+	}
+
 	// Check if Status is empty and status annotation is present
 	// Manually restore data.
 	if !r.hostHasStatus(host) {
@@ -370,6 +400,33 @@ func (r *ReconcileBareMetalHost) credentialsErrorResult(err error, request recon
 	}
 }
 
+// hasRebootAnnotation checks for existence of reboot annotations and returns true if at least one exist
+func hasRebootAnnotation(host *metal3v1alpha1.BareMetalHost) bool {
+	for annotation := range host.Annotations {
+		if isRebootAnnotation(annotation) {
+			return true
+		}
+	}
+	return false
+}
+
+// isRebootAnnotation returns true if the provided annotation is a reboot annotation (either suffixed or not)
+func isRebootAnnotation(annotation string) bool {
+	return strings.HasPrefix(annotation, rebootAnnotationPrefix+"/") || annotation == rebootAnnotationPrefix
+}
+
+// clearRebootAnnotations deletes all reboot annotations exist on the provided host
+func clearRebootAnnotations(host *metal3v1alpha1.BareMetalHost) (dirty bool) {
+	for annotation := range host.Annotations {
+		if isRebootAnnotation(annotation) {
+			delete(host.Annotations, annotation)
+			dirty = true
+		}
+	}
+
+	return
+}
+
 // Manage deletion of the host
 func (r *ReconcileBareMetalHost) actionDeleting(prov provisioner.Provisioner, info *reconcileInfo) actionResult {
 	info.log.Info(
@@ -526,28 +583,21 @@ func (r *ReconcileBareMetalHost) actionMatchProfile(prov provisioner.Provisioner
 
 // Start/continue provisioning if we need to.
 func (r *ReconcileBareMetalHost) actionProvisioning(prov provisioner.Provisioner, info *reconcileInfo) actionResult {
-	getUserData := func() (string, error) {
-		if info.host.Spec.UserData == nil {
-			info.log.Info("no user data for host")
-			return "", nil
-		}
-		info.log.Info("fetching user data before provisioning")
-		userDataSecret := &corev1.Secret{}
-		key := types.NamespacedName{
-			Name:      info.host.Spec.UserData.Name,
-			Namespace: info.host.Spec.UserData.Namespace,
-		}
-		err := r.client.Get(context.TODO(), key, userDataSecret)
-		if err != nil {
-			return "", errors.Wrap(err,
-				"failed to fetch user data from secret reference")
-		}
-		return string(userDataSecret.Data["userData"]), nil
+	hostConf := &hostConfigData{
+		host:   info.host,
+		log:    info.log,
+		client: r.client,
 	}
-
 	info.log.Info("provisioning")
 
-	provResult, err := prov.Provision(getUserData)
+	if clearRebootAnnotations(info.host) {
+		if err := r.client.Update(context.TODO(), info.host); err != nil {
+			return actionError{errors.Wrap(err, "failed to remove reboot annotations from host")}
+		}
+		return actionContinueNoWrite{}
+	}
+
+	provResult, err := prov.Provision(hostConf)
 	if err != nil {
 		return actionError{errors.Wrap(err, "failed to provision")}
 	}
@@ -593,6 +643,13 @@ func (r *ReconcileBareMetalHost) actionDeprovisioning(prov provisioner.Provision
 		return actionContinue{provResult.RequeueAfter}
 	}
 
+	if clearRebootAnnotations(info.host) {
+		if err = r.client.Update(context.TODO(), info.host); err != nil {
+			return actionError{errors.Wrap(err, "failed to remove reboot annotations from host")}
+		}
+		return actionContinueNoWrite{}
+	}
+
 	// After the provisioner is done, clear the image settings so we
 	// transition to the next state.
 	info.host.Status.Provisioning.Image = metal3v1alpha1.Image{}
@@ -619,19 +676,40 @@ func (r *ReconcileBareMetalHost) manageHostPower(prov provisioner.Provisioner, i
 		return actionContinue{provResult.RequeueAfter}
 	}
 
+	desiredPowerOnState := info.host.Spec.Online
+
+	if !info.host.Status.PoweredOn {
+		if _, suffixlessAnnotationExists := info.host.Annotations[rebootAnnotationPrefix]; suffixlessAnnotationExists {
+			delete(info.host.Annotations, rebootAnnotationPrefix)
+
+			if err = r.client.Update(context.TODO(), info.host); err != nil {
+				return actionError{errors.Wrap(err, "failed to remove reboot annotation from host")}
+			}
+
+			return actionContinueNoWrite{}
+		}
+	}
+
+	provState := info.host.Status.Provisioning.State
+	isProvisioned := provState == metal3v1alpha1.StateProvisioned || provState == metal3v1alpha1.StateExternallyProvisioned
+	if hasRebootAnnotation(info.host) && isProvisioned {
+		desiredPowerOnState = false
+	}
+
 	// Power state needs to be monitored regularly, so if we leave
 	// this function without an error we always want to requeue after
 	// a delay.
 	steadyStateResult := actionContinue{time.Second * 60}
-	if info.host.Status.PoweredOn == info.host.Spec.Online {
+	if info.host.Status.PoweredOn == desiredPowerOnState {
 		return steadyStateResult
 	}
 
 	info.log.Info("power state change needed",
-		"expected", info.host.Spec.Online,
-		"actual", info.host.Status.PoweredOn)
+		"expected", desiredPowerOnState,
+		"actual", info.host.Status.PoweredOn,
+		"reboot process", desiredPowerOnState != info.host.Spec.Online)
 
-	if info.host.Spec.Online {
+	if desiredPowerOnState {
 		provResult, err = prov.PowerOn()
 	} else {
 		provResult, err = prov.PowerOff()
@@ -647,7 +725,7 @@ func (r *ReconcileBareMetalHost) manageHostPower(prov provisioner.Provisioner, i
 	if provResult.Dirty {
 		info.postSaveCallbacks = append(info.postSaveCallbacks, func() {
 			metricLabels := hostMetricLabels(info.request)
-			if info.host.Spec.Online {
+			if desiredPowerOnState {
 				metricLabels[labelPowerOnOff] = "on"
 			} else {
 				metricLabels[labelPowerOnOff] = "off"
